/*
hopen.c Version 0.1.0  Hash open database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define HOPEN 200

hashfmt *hopen(char *dbname, int rwflag,
   int *reason)
   {
   BOOL bResult;
   DWORD msgerr;
   LPVOID lpMsgBuf;        
   DWORD rdlen;
   int namelen;
   HANDLE hndl;
   unsigned char *p,*q;
   hashfmt *hshhdr;
   *reason = 0;
   namelen = strlen(dbname);
   if (namelen < 1 || namelen > 128)
      {
      fprintf(stderr,"hopen: wrong "
         "length name %s\n",
	 dbname);
      *reason = HASH_ERR + HOPEN + 9;
      return(NULL);
      } /* dbname wrong length */
   if (rwflag == UPDT)
      {
      hndl = CreateFile(dbname,
         GENERIC_READ | GENERIC_WRITE,
         0,              /* do not share */
         NULL,           /* cannot be inherited */
         OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
         NULL);          /* no template */
      } /* if update */
   else if (rwflag == READ)
      {
      hndl = CreateFile(dbname,
         GENERIC_READ,
         0,              /* do not share */
         NULL,           /* cannot be inherited */
         OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
         NULL);          /* no template */
      } /* if read only */
   else
      {
      fprintf(stderr,"hopen: invalid read/write flag\n");
      *reason = HASH_ERR + HOPEN + 1;
      return(NULL);
      } /* rwflag err */
   if (hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      *reason = (int) SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"hopen: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hopen: %s: open ", dbname);
      if (rwflag == READ)
         fprintf(stderr,"read-only\n");
      else if (rwflag == UPDT)
         fprintf(stderr,"read/write\n");
      return(NULL);
      } /* if open error */
   hshhdr = (hashfmt *) malloc(HDRLEN+HDRLEN);
   if (hshhdr == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr\n");
      bResult = CloseHandle(hndl);
      *reason = HASH_ERR + HOPEN + 3;
      return(NULL);
      } /* out of mem */
   bResult = ReadFile(hndl,hshhdr,HDRLEN,
      &rdlen,NULL);
   if (!bResult)
      {
      msgerr = GetLastError();
      *reason = (int) SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"hopen: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hopen: dbname %s\n", dbname);
      fprintf(stderr,"hopen: error "
         "reading header\n");
      bResult = CloseHandle(hndl);
      free(hshhdr);
      return(NULL);
      } /* read err */
   if (rdlen != HDRLEN)
      {
      fprintf(stderr,"hopen: dbname %s\n", dbname);
      fprintf(stderr,"hopen: error "
         "reading header\n");
      bResult = CloseHandle(hndl);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 4;
      return(NULL);
      } /* read err */
   if (strncmp(hshhdr->signature,
      H_SIGNATURE,sizeof(H_SIGNATURE)))
      {
      fprintf(stderr,"hopen: dbname %s\n", dbname);
      fprintf(stderr,"hopen: error "
	 "opening non-hash file\n");
      bResult = CloseHandle(hndl);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 5;
      return(NULL);
      } /* invalid signature */
   p = (unsigned char *) hshhdr + HDRLEN;
   q = (unsigned char *) p + HDRLEN;
   while (p < q) *p++ = '\0';
   hshhdr->hndl = hndl;

   hshhdr->rcdnode = hshhdr->rcdhdr + hshhdr->rcdlen;

   { /* calculate offset of freestk */
   int diff;
   hashfmt tmphdr;
   p = (unsigned char *) &tmphdr.freestk;
   q = (unsigned char *) &tmphdr.signature[0];
   diff = (unsigned int) p - (unsigned int) q;
   hshhdr->fsofst  = (long) diff;
   } /* calculate offset of freestk */

   hshhdr->hashend = (long) HDRLEN
      + (hshhdr->totrcds * hshhdr->rcdnode);

   hshhdr->node = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->node == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf1\n");
      hclose(hshhdr);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 6;
      return(NULL);
      } /* out of mem */
   hshhdr->rcd = (unsigned char *)
      hshhdr->node + hshhdr->rcdhdr;
   hshhdr->key = (unsigned char *)
      hshhdr->rcd + hshhdr->keyofst;
   hshhdr->nodenext = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->nodenext == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf2\n");
      hclose(hshhdr);
      free(hshhdr->node);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 7;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) hshhdr->nodenext;
   q = (unsigned char *) p + hshhdr->rcdnode;
   while (p < q) *p++ = '\0';
   hshhdr->nodeprev = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->nodeprev == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf3\n");
      hclose(hshhdr);
      free(hshhdr->node);
      free(hshhdr->nodenext);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 8;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) hshhdr->nodeprev;
   q = (unsigned char *) p + hshhdr->rcdnode;
   while (p < q) *p++ = '\0';
   strcpy(hshhdr->dbname,dbname);
   return(hshhdr);
   } /* hopen */
